#include "Funciones.h"//Cabecera con las declaraciones 
#include <Arduino.h>//Incluye la librería para emplear funciones propias de arduino
#include <LiquidCrystal_I2C.h>//Incluye la librería para LCD controlada por I2C

//SENSORES
int Tad590=0;//temperatura AD590
int Tntc=0; //temperatura NTC
int Tpt100=0;//temperatura PT100
int Ttermopar=0;//temperatura Termopar

//FILTROS
const int ventana=10;

int bufferCircularAD[ventana];
int sumaBufferCircularAD=0;
int indiceAD=0;

int bufferCircularNTC[ventana];
int sumaBufferCircularNTC=0;
int indiceNTC=0;

int bufferCircularPT[ventana];
int sumaBufferCircularPT=0;
int indicePT=0;

int bufferCircularK[ventana];
int sumaBufferCircularK=0;
int indiceK=0;

//ENTRADAS
int Modo=1; //1=Manual 0=PC  
int Sensor=1; //1=AD590 2=Termopar 3=PT100 4=NTC
int Ventilador=1; //1=OFF 0=ON
int VelVentilador=255;

//FLAGS
int inicio_PC=0; //Flag para detectar el cambio de modo
int Tmuestreo=0; //Flag tiempo de muestreo 

//CONTROL PID
int Consigna=0,Kp=0,Ki=0,Kd=0;
int ErrorActual=0;
int ErrorAnterior=0;
float iError=0;
float dError=0;
int pTerm=0;
int iTerm=0;
int dTerm=0;
int LimInt=0;
int Temperatura=0; 
int Salida_PID=0;
const int Salida_PID_max=255;
float Tm=0.5;

//RESISTENCIA PWM
int DC=0;
int Duty=0;

//LCD
LiquidCrystal_I2C lcd(0x3F,20,4);

//INTERRUPCIÓN
//Configura el Timer 4 para interrupción cada 500ms
void configTimer4Interrupt(){
  cli(); //Desactiva interrupciones
  TCCR4A = (0<<COM4A1)|(0<<COM4A0)|(0<<COM4B1)|(0<<COM4B0)|(0<<COM4C1)|(0<<COM4C0)|(0<<WGM41)|(0<<WGM40);
  //Divisor de frecuencia 64 y modo CTC
  TCCR4B = (0<<ICNC4)|(0<<ICES4)|(0<<5)|(0<<WGM43)|(1<<WGM42)|(0<<CS42)|(1<<CS41)|(1<<CS40); 
  OCR4A = 62499; // Límite 500ms
  TCNT4 = 0;// Resetea el contador
  TIMSK4 = (0<<ICIE4)|(0<<OCIE4B)|(1<<OCIE4A)|(0<<TOIE4); //Interrupción por límite de cuenta  
  sei(); //Activa interrupciones
}

//Función a ejecutar tras interrupción 500ms
ISR(TIMER4_COMPA_vect) { 
  //Detiene el contador
  TCCR4B = (0<<ICNC4)|(0<<ICES4)|(0<<5)|(0<<WGM43)|(1<<WGM42)|(0<<CS42)|(0<<CS41)|(0<<CS40);
  TIMSK4 = (0<<ICIE4)|(0<<OCIE4B)|(0<<OCIE4A)|(0<<TOIE4);//Deshabilita la interrupción
  Tmuestreo=1; //Flag tiempo de muestreo
  TCNT4 = 0; //Resetea el contador
  //Activa el contador
  TCCR4B = (0<<ICNC4)|(0<<ICES4)|(0<<5)|(0<<WGM43)|(1<<WGM42)|(0<<CS42)|(1<<CS41)|(1<<CS40);
  TIMSK4 = (0<<ICIE4)|(0<<OCIE4B)|(1<<OCIE4A)|(0<<TOIE4);//Habilita la interrupción
}

//SENSORES
void FiltradoAD(){ //Filtrado de la temperatura del AD590
   sumaBufferCircularAD-=bufferCircularAD[indiceAD]; //Resta la lectura más antigua
   bufferCircularAD[indiceAD]=Tad590; //sustituye la lectura más antigua por la actual 
   sumaBufferCircularAD+=bufferCircularAD[indiceAD]; //Actualiza la suma
   indiceAD++; //Actualiza el índice
   if(indiceAD==ventana){ //Valor máximo del índice alcanzado
    indiceAD=0; //Resetea el índice
   }
   //El valor filtrado es la media de los valores del buffer
   Tad590=sumaBufferCircularAD/ventana; 
}

void Calc_Temp_ad590(int pinAD590){
  int AD590; //valor digital AD590
  float Vad590; //tensión 
  AD590=analogRead(pinAD590);//lee valor digital del pin AD590
  Vad590=(float(AD590)*5)/1023; //conversión del valor digital a tensión AD590
  Tad590=(Vad590/0.011)-273; //obtiene la temperatura del AD590
}

void FiltradoNTC(){
  sumaBufferCircularNTC-=bufferCircularNTC[indiceNTC];
  bufferCircularNTC[indiceNTC]=Tntc;
  sumaBufferCircularNTC+=bufferCircularNTC[indiceNTC];
  indiceNTC++;
  if(indiceNTC==ventana){
   indiceNTC=0;
  }
  Tntc=sumaBufferCircularNTC/ventana;
}

void Calc_Temp_ntc(int pinNTC){
  int NTC; //lectura valor digital NTC
  float Vntc; //tensión
  NTC=analogRead(pinNTC);//lee valor digital NTC
  Vntc=(float(NTC)*5)/1023; //conversión del valor digital a tensión NTC
  //obtiene la temperatura de la ntc
  Tntc=((7.08*Vntc+6.606)/(0.0152639*Vntc+0.029764605))-273; 
}

void FiltradoPT(){
   sumaBufferCircularPT-=bufferCircularPT[indicePT];
   bufferCircularPT[indicePT]=Tpt100;
   sumaBufferCircularPT+=bufferCircularPT[indicePT];
   indicePT++;
   if(indicePT==ventana){
    indicePT=0;
   }
   Tpt100=sumaBufferCircularPT/ventana;
}

void Calc_Temp_pt100(int pinPT100){
  int PT100; //lectura valor digital PT100
  float Vpt100; //tensión
  PT100=analogRead(pinPT100);//lee valor digital PT100
  Vpt100=(float(PT100)*5)/1023; //conversión del valor digital a tensión PT100
  Tpt100=(5.04-Vpt100)/(0.033341); //obtiene la temperatura de la PT100
}

void FiltradoK(){
   sumaBufferCircularK-=bufferCircularK[indiceK];
   bufferCircularK[indiceK]=Ttermopar;
   sumaBufferCircularK+=bufferCircularK[indiceK];
   indiceK++;
   if(indiceK==ventana){
    indiceK=0;
   }
   Ttermopar=sumaBufferCircularK/ventana;
}

void Calc_Temp_termopar(int pinTermopar){
  int Termopar; //valor digital Termopar
  float Vtermopar; //tensión
  Termopar=analogRead(pinTermopar);//lee valor digital Termopar
  //conversión del valor digital a tensión Termopar
  Vtermopar=(float(Termopar)*5)/1023; 
  Ttermopar=100*Vtermopar; //obtiene la temperatura del Termopar
}

void initFiltros(){
  for(int i=0;i<ventana;i++){ //Recorre los buffers 
    Calc_Temp_ad590(pinAD590); //Obtiene la temperatura del sensor
    bufferCircularAD[i]=Tad590; //La almacena en el buffer
    sumaBufferCircularAD+=bufferCircularAD[i]; //Actualiza la suma
    Calc_Temp_ntc(pinNTC);
    bufferCircularNTC[i]=Tntc;
    sumaBufferCircularNTC+=bufferCircularNTC[i];
    Calc_Temp_pt100(pinPT100);
    bufferCircularPT[i]=Tpt100;
    sumaBufferCircularPT+=bufferCircularPT[i];
    Calc_Temp_termopar(pinTermopar);
    bufferCircularK[i]=Ttermopar;
    sumaBufferCircularK+=bufferCircularK[i];
    delay(10); //Espera 10ms entre muestras
  }
}
//ENTRADAS
void Lect_Entradas_Manual(){
  if(digitalRead(pinSelec_AD590)==LOW){ //Lectura del selector
    Sensor=1; //Asignación del número correspondiente al sensor
  }else if(digitalRead(pinSelec_Termopar)==LOW){
    Sensor=2;
  }else if(digitalRead(pinSelec_PT100)==LOW){
    Sensor=3;
  }else if(digitalRead(pinSelec_NTC)==LOW){
    Sensor=4;
  }
  Consigna=analogRead(pinConsigna); //Lectura potenciómetro Consigna
  Consigna=map(Consigna,0,1023,0,150); //Ajusta el valor digital al rango 0-150ºC
  //Lectura potenciómetros Kp, Ki y Kd y ajuste de rango
  Kp=analogRead(pinKp);
  Kp=map(Kp,0,1023,0,100);
  Ki=analogRead(pinKi);
  Ki=map(Ki,0,1023,0,100);
  Kd=analogRead(pinKd);
  Kd=map(Kd,0,1023,0,50);
  Ventilador=digitalRead(pinIntVentilador); //Lectura interruptor Ventilador
}
//Almacena los 6 bytes recibidos en el buffer de entrada
void Lect_Entradas_PC(){ 
    Sensor=Serial.read();
    Consigna=Serial.read();
    Kp=Serial.read();
    Ki=Serial.read();
    Kd=Serial.read();
    Ventilador=Serial.read();
}

//Escritura de los 5 bytes a enviar en el buffer de salida
void Enviar_Datos_PC(){ 
  Serial.write(Tad590); 
  Serial.write(Ttermopar);
  Serial.write(Tpt100);
  Serial.write(Tntc);
  Serial.write(Duty);
}

//VENTILADOR
void Control_Ventilador(){
  if(Ventilador==1){ //Ventilador OFF
    analogWrite(pinPWMVentilador,0); //Apaga el ventilador
  }else{ //Ventilador ON
    //Lectura de la consigna de velocidad
    VelVentilador=analogRead(pinVelVentilador); 
    //Ajuste de la precisión de  10 bits a 8 bits
    VelVentilador=map(VelVentilador,0,1023,0,255); 
    //Control de velocidad según consigna
    analogWrite(pinPWMVentilador,VelVentilador); 
  }
}

//CONTROL PID
void Params_PID(){
  switch(Sensor){  //Asigna la temperatura de referencia según sensor seleccionado
    case 1:
      Temperatura=Tad590;
      break;
    case 2:
      Temperatura=Ttermopar;
      break;
    case 3:
      Temperatura=Tpt100;
      break;
    case 4:
      Temperatura=Tntc;
      break;
    default:
      Temperatura=Tad590;
  }
}

void Control_PID(){
  Params_PID();
  ErrorAnterior=ErrorActual; //Almacena error anterior
  ErrorActual=(Consigna-Temperatura); //Calcula error actual
  //Término Proporcional
  pTerm=Kp*ErrorActual; //Calcula el término proporcional
  //Término Integral
  if(abs(ErrorActual)>LimInt ) { //Error mayor que el límite                                   
    iError=((float(ErrorActual)+float(ErrorAnterior))/2)*Tm; //Calcula error integral
    iTerm=Ki*iError; //Calcula el término integral
  }
  else {
    iTerm = 0; //No se aplica acción integral
  }
  if (iTerm < 0){
    iTerm = 0; //Mínimo valor del término integral
  }
  //Término Derivativo
  dError=(float(ErrorActual)-float(ErrorAnterior))/Tm; //Calcula error derivativo
  dTerm=Kd*dError; //Calcula el término derivativo
  if(dTerm<0){
    dTerm=0; //Mínimo valor del término derivativo
  }
  //Acción PID
  Salida_PID=pTerm+iTerm+dTerm; //Cálculo de la señal PID  
  if (Salida_PID > Salida_PID_max){
    Salida_PID = Salida_PID_max; //Máximo valor de la salida
  }
  if (Salida_PID < 0){
    Salida_PID = 0; //Mínimo valor de la salida
  }
  Salida_PID=map(Salida_PID,0,Salida_PID_max,0,16000); //Ajuste del rango 
}


//RESISTENCIA PWM
void configTimerPWM(){
  //Salida PWM en OC5A 
  TCCR5A=(1<<COM5A1)|(0<<COM5A0)|(0<<COM5B1)|(0<<COM5B0)|(0<<COM5C1)|(0<<COM5C0)|(0<<WGM51)|(0<<WGM50);
  //Modo control de fase y frecuencia, prescaler 1
  TCCR5B=(0<<ICNC5)|(0<<ICES5)|(0<<5)|(1<<WGM53)|(0<<WGM52)|(0<<CS52)|(0<<CS51)|(1<<CS50);
  ICR5=16000; //Frecuencia PWM 500Hz
  TCNT5=0;//Resetea Timer 5
  OCR5A=16000;//registro para controlar el ciclo de trabajo
}

void Resistencia_PWM(){
  DC=map(Salida_PID,0,16000,16000,0);//Invierte la salida
  Duty=100-(float(DC)/16000)*100; //Duty Cycle en %
  OCR5A=DC; //Actualiza el registro del contador 5 
}

void AlertaEnfriamiento(){
  OCR5A=16000; //Desconecta la resistencia calefactora
  analogWrite(pinPWMVentilador,255); //Enciende el ventilador a su velocidad máxima
  //Muestra mentase de alerta en pantalla y temperatura del horno
  lcd.setCursor(0,0);
  lcd.print("                    ");
  lcd.setCursor(0,1);
  lcd.print("ALERTA ENFRIAMIENTO "); 
  lcd.setCursor(0,2);
  lcd.print("      TEMP:         ");
  lcd.setCursor(0,3);
  lcd.print("                    ");
  do{
    Calc_Temp_ad590(pinAD590); //Calcula la temperatura
    lcd.setCursor(11,2);
    lcd.print(Tad590); //Imprime la temperatura
    lcd.print("C");
    lcd.setCursor(11,2);
    lcd.print("     ");
  }while(Tad590 > 100); 
  //El bucle finaliza cuando la temperatura es igual o inferior a 100C
  analogWrite(pinPWMVentilador,0); //Apaga el ventilador
  //Reset pantalla a estado anterior
  lcd.clear();
  imprimir_titulos();
}

//LCD
void init_lcd(){
  lcd.begin();// Inicializar el display LCD
  lcd.backlight();//Encender la luz de fondo
}

void imprimir_saludo(){ //Imprime mensaje de inicio
  lcd.setCursor(7,1); //Coloca el cursor
  //Imprime mensaje
  lcd.print("CONTROL"); 
  lcd.setCursor(3,2);
  lcd.print("DE TEMPERATURA");
  delay(3000); //Espera 3s
  lcd.clear(); //Limpia la pantalla 
}

void imprimir_modo(){ //Imprime el modo de funcionamiento
  lcd.setCursor(0,3);
  lcd.print("           "); //Borra modo anterior
  lcd.setCursor(0,3);
  if(Modo==1){  //Modo manual activo
    lcd.print("/manual/");
  }else{ //Modo PC activo
    lcd.print("/pc/");
  }
}
void imprimir_titulos(){ //Imprime los nombres de las variables
  lcd.setCursor(0,1);
  lcd.print("Consig:");
  lcd.setCursor(0,2);
  lcd.print("Error:");
  lcd.setCursor(11,0);
  lcd.print("Kp:");
  lcd.setCursor(11,1);
  lcd.print("Ki:");
  lcd.setCursor(11,2);
  lcd.print("Kd:");
  lcd.setCursor(11,3);
  lcd.print("DC:");
}
//Imprime el nombre del sensor seleccionado y su temperatura
void imprimir_temp(){ 
  lcd.setCursor(0,0);
  lcd.print("           ");
  lcd.setCursor(0,0);
  switch(Sensor){ 
    case 1:
      lcd.print("AD590:");
      break;
    case 2:
      lcd.print("TIPOK:");
      break;
    case 3:
      lcd.print("PT100:");
      break;
    case 4:
      lcd.print("NTC:");
      break;
    default:
      lcd.print("AD590:");
  }
  lcd.print(Temperatura);
  lcd.print("C");
}
void imprimir_consig(){ //Imprime el valor de consigna
  lcd.setCursor(7,1);
  lcd.print("    ");
  lcd.setCursor(7,1);
  lcd.print(Consigna);
  lcd.print("C");
}

void imprimir_error(){ //Imprime el valor del error actual
  lcd.setCursor(6,2);
  lcd.print("     ");
  lcd.setCursor(6,2);
  lcd.print(ErrorActual);
  lcd.print("C");
}

void imprimir_Kp(){ //Imprime el valor de la constante Kp
  lcd.setCursor(14,0);
  lcd.print("      ");
  lcd.setCursor(14,0);
  lcd.print(Kp);
}

void imprimir_Ki(){ //Imprime el valor de la constante Ki
  lcd.setCursor(14,1);
  lcd.print("      ");
  lcd.setCursor(14,1);
  lcd.print(Ki);
}
void imprimir_Kd(){ //Imprime el valor de la constante Kd
  lcd.setCursor(14,2);
  lcd.print("      ");
  lcd.setCursor(14,2);
  lcd.print(Kd);
}

void imprimir_DC(){ //Imprime el valor del ciclo de trabajo
  lcd.setCursor(14,3);
  lcd.print("      ");
  lcd.setCursor(14,3);
  lcd.print(Duty);
}
